#pragma once

#include <memory>
#include <type_traits>
#include <cassert>

template< typename T, size_t _block_size = 256  >
struct link_n_list
{
    using value_type = T;
    using reference = T &;
    using const_reference = const T &;
    using difference_type = ptrdiff_t;
    using size_type = size_t;

    static constexpr size_t block_size = _block_size;

  private:
    struct Node
    {
        Node() : _end( begin() ) { }
        Node( const Node &o, Node *prev ) :
            _end( begin() + o.size() ), _prev( prev )
        {
            std::uninitialized_copy( o.begin(), o.end(), begin() );
        }
        ~Node() {
            std::destroy( begin(), end() );
        }

        size_type size() const { return end() - begin(); }

        bool full() const {
            return _end == begin() + block_size;
        }

        bool empty() const {
            return _end == begin();
        }

        // note: reinterpret_cast from char pointer is one of the few OK ones
        value_type *begin() { return reinterpret_cast< value_type * >( _data ); } // NOLINT
        const value_type *begin() const { return reinterpret_cast< const value_type * >( _data ); } // NOLINT

        value_type *end() { return _end; }
        const value_type *end() const { return _end; }

        template< typename TT >
        reference push_back( TT &&x ) {
            assert( !full() );
            // inplace new: construct element in given, already allocated, space
            new ( _end ) value_type( std::forward< TT >( x ) );
            ++_end;
            return _end[-1];
        }

        // forward pointers are unique_ptr, back has to be raw to prevent
        // circular structure of unique_ptrs
        std::unique_ptr< Node > _next;
        value_type *_end;
        // alignas( alignof( value_type ) ) ensures data are properly aligned
        alignas( alignof( value_type ) ) char _data[ block_size * sizeof( value_type ) ];
        Node *_prev = nullptr;
    };

    // a tag to indicate if we are creating end iterator
    struct End { };

    // we parametrize the iterator by either Node * or const Node * and
    // a corresponding value_type * or const value_type * to create const and
    // non const iterator from the same code
    template< typename NodePtr, typename Ptr >
    struct Iterator {
        Iterator() = default;
        explicit Iterator( NodePtr n ) : _node( n ), _ptr( n->begin() ) { }
        Iterator( NodePtr n, End /* unused */ ) : _node( n ), _ptr( n->end() ) { }

        // allow casting iterator to const_iterator (for the reverse way it
        // will fail to compile)
        template< typename N, typename P >
        Iterator( const Iterator< N, P > &o ) : _node( o._node ), _ptr( o._ptr ) { } // NOLINT


        // these all must be defined for ForwardIterator
        using value_type = link_n_list::value_type;
        // to have less template arguments derive reference type from pointer
        using reference = std::add_lvalue_reference_t< std::remove_pointer_t< Ptr > >;
        using pointer = Ptr;
        using difference_type = link_n_list::difference_type;
        // specifies the kind of the iterator
        using iterator_category = std::forward_iterator_tag;

        template< typename N, typename P >
        bool operator==( const Iterator<N, P> &o ) const { return _ptr == o._ptr; }
        template< typename N, typename P >
        bool operator!=( const Iterator<N, P> &o ) const { return !(*this == o); }

        Iterator &operator++() {
            // note: we first increment _ptr and then check if it is at the end,
            // this ensures that we don't need to handle the end of the entire list
            // as a special case if the last node is full
            ++_ptr;
            if ( _ptr == _node->end() && _node->_next ) {
                _node = _node->_next.get();
                _ptr = _node->begin();
            }
            return *this;
        }

        Iterator operator++( int ) {
            auto copy{ *this };
            ++*this;
            return copy;
        }

        reference operator*() const { return *_ptr; }
        pointer operator->() const { return _ptr; }

      private:
        // one of these friends is us, but the compiler does not mind
        friend Iterator< Node *, value_type * >;
        friend Iterator< const Node *, const value_type * >;
        NodePtr _node = nullptr;
        Ptr _ptr = nullptr;
    };
  public:

    using iterator = Iterator< Node *, value_type * >;
    using const_iterator = Iterator< const Node *, const value_type * >;

    link_n_list() : _first( std::make_unique< Node >() ), _last( _first.get() ) { }
    // no need for explicit destructor as unique_ptrs take care of deallocation

    // copying is tricky with unique_ptrs, we have to build new linked list
    // based on the original, copying nodes is no good (and should not be
    // allowed)
    link_n_list( const link_n_list &o ) {
        if ( o.empty() )
            return;

        const Node *ptr = o._first.get();
        _first = std::make_unique< Node >( *ptr, nullptr );
        Node *prev = _first.get();
        ptr = ptr->_next.get();

        while ( ptr ) {
            prev->_next = std::make_unique< Node >( *ptr, prev );
            prev = prev->_next.get();
            ptr = ptr->_next.get();
        }
        assert( prev );
        _last = prev;
    }
    link_n_list( link_n_list &&o ) noexcept = default;

    link_n_list &operator=( link_n_list o ) {
        swap(o);
    }

    void swap( link_n_list &o ) {
        using std::swap;
        swap( _first, o._first );
        swap( _last, o._last );
    }

    // unique_ptr has only explicit cast to bool, so either compare it with
    // nullptr, or use explicit cast such as bool( _first )
    bool empty() const { return _first->empty(); }

    iterator begin() { return iterator( _first.get() ); }
    const_iterator begin() const { return const_iterator( _first.get() ); }
    const_iterator cbegin() const { return const_iterator( _first.get() ); }

    iterator end() { return iterator( _last, End() ); }
    const_iterator end() const { return const_iterator( _last, End() ); }
    const_iterator cend() const { return const_iterator( _last, End() ); }

    reference front() { return _first->begin()[0]; }
    const_reference front() const { return _first->begin()[0]; }

    reference back() { return _last->end()[-1]; }
    const_reference back() const { return _last->end()[-1]; }

    // iterators are never invalidated by push_back
    reference push_back( const value_type &v ) { return _push_back( v ); }
    reference push_back( value_type &&v ) { return _push_back( std::move( v ) ); }

  private:
    template< typename TT >
    reference _push_back( TT &&val ) {
        if ( !_last ) {
            _first = std::make_unique< Node >();
            _last = _first.get();
        }
        else if ( _last->full() ) {
            _last->_next = std::make_unique< Node >();
            _last->_next->_prev = _last;
            _last = _last->_next.get();
        }
        return _last->push_back( std::forward< TT >( val ) );
    }

    // first is kept using unique_ptr to allow automatic deallocation, last has
    // to be raw pointer to prevent double ownership
    std::unique_ptr< Node > _first;
    Node *_last;
};
